(* Les graphes seront représentés sous forme de listes d'adjacence *)
type sommet = int ;;
type graphe = ( sommet list ) array ;;
type chemin = ( sommet list ) ;;
type cycle = ( sommet list ) ;;

let recherche_cycle ( g : graphe ) : cycle =
  let visite = Array.make ( Array.length g ) false in
  let rec extraire_cycle
            ( historique : chemin )
            ( depart : sommet )
            ( cycle : cycle ): cycle =
    match historique with
    | [] -> failwith "Aucun cycle trouvé."
    | i :: hist when i = depart -> depart :: cycle
    | i :: hist -> extraire_cycle hist depart ( i :: cycle ) in
  let rec parcours_composante
            ( noeud : sommet ) ( historique : chemin  ) : cycle =
    if visite.(noeud) then extraire_cycle historique noeud [noeud]
    else
      let rec parcours_voisins ( voisins : sommet list ) : cycle =
        match voisins, historique with
          | [], _ -> []
          | v :: voisins, parent :: _ when v = parent ->
            parcours_voisins voisins
          | v :: voisins, _ ->
             let cycle = parcours_composante v ( noeud :: historique ) in
             if cycle = [] then parcours_voisins voisins
             else cycle in
      begin
        visite.(noeud) <- true;
        parcours_voisins g.(noeud)
      end in
  let rec parcours ( i : sommet ) : cycle =
    if i = Array.length g then []
    else if visite.(i) then parcours ( i + 1 )
    else let cycle = parcours_composante i [] in
         if cycle = [] then parcours ( i + 1 ) else cycle in
  parcours 0
;;

let nombre_arete ( g : graphe ) : int =
  let nb = ref 0 in
  for i = 0 to ( Array.length g ) - 1 do
    nb := !nb + ( List.length g.(i) )
  done ;
  !nb / 2
;;

(* Prérequis : c est un cycle de g *)
let est_eulerien ( g : graphe ) ( c : cycle ) : bool =
  let n = Array.length g in
  let memo = Array.make_matrix n n false in
  let rec parcours_arete ( c : chemin ) : bool =
    match c with
    | [_] | [] -> true
    | s :: s' :: chemin when memo.(s).(s') -> false
    | s :: s' :: chemin -> begin
        memo.(s).(s') <- true;
        memo.(s').(s) <- true;
        parcours_arete ( s' :: chemin ) end in
  ( nombre_arete g ) = ( List.length c) - 1 && parcours_arete c
;;

let supprime_cycle ( g : graphe ) : cycle =
  let cycle = recherche_cycle g in
  let rec remove a = function
    | [] -> []
    | i :: l when i = a -> l
    | i :: l -> i :: (remove a l) in
  let elimine_arete ( s : sommet ) ( s' : sommet ) : unit =
    g.(s) <- remove s' g.(s);
    g.(s') <- remove s g.(s') in
  let rec elimine_chaque_arete ( c : chemin ) : unit =
    match c with
    | [_] | [] -> ()
    | s :: s' :: chemin -> ( elimine_arete s s';
                             elimine_chaque_arete ( s' :: chemin ) ) in
  elimine_chaque_arete cycle;
  cycle;
;;

let a_cycle_eulerien ( g : graphe ) : bool =
  let degre ( g : graphe ) ( s : sommet ) : int = List.length g.(s) in
  let rec parcours ( i : sommet ) : bool =
    if i = Array.length g then true
    else if ( degre g i ) mod 2 = 0 then parcours ( i + 1 )
    else false in
  parcours 0
;;

let cycle_eulerien ( g : graphe ) : cycle =
  if not ( a_cycle_eulerien g ) then []
  else begin
    let g' = Array.make ( Array.length g ) [] in
    for i = 0 to ( Array.length g ) - 1 do
        g'.(i) <- g.(i)
    done ;
    let cycles = ref [] in
    while nombre_arete g' != 0 do
        cycles := ( supprime_cycle g' ) :: !cycles
    done ;

let rotation_cycle ( s : sommet ) ( c : cycle ) : cycle =
    let rec apres_s = function
    | [] -> failwith "Sommet hors du cycle"
    | i :: chemin when i = s -> s :: chemin
    | _ :: chemin -> apres_s chemin in
    let rec avant_s = function
    | [] -> failwith "Sommet hors du cycle"
    | i :: chemin when i = s -> [s]
    | i :: chemin -> i :: ( avant_s chemin ) in
    match c with
    | [] -> failwith "Cycle Invalide"
    | s' :: _ when s' = s -> c
    | s' :: chemin -> ( apres_s chemin ) @ ( avant_s chemin ) in

let fusion ( c1 : cycle ) ( c2 : cycle ) ( s : sommet ) : cycle =
    let rot_c1 = rotation_cycle s c1 in
    let rot_c2 = rotation_cycle s c2 in
    match rot_c2 with
    | [] -> failwith "Cycle invalide"
    | s2 :: chemin -> rot_c1 @ chemin in

let recherche_sommet_commun ( c1 : cycle ) ( c2 : cycle ) : sommet =
    let rec parcours_cycle ( c : cycle ) ( c' : cycle ) : sommet =
    match c, c' with
    | [], _ -> -1
    | s1 :: c, [] -> parcours_cycle c c2
    | s1 :: _, s2 :: _ when s1 = s2 -> s1
    | _, _ :: c' -> parcours_cycle c c' in
    parcours_cycle c1 c2 in

let fusion_liste ( cycles : cycle list ) : cycle =
    let rec iterations
            ( c_euler : cycle )
            ( cycles1 : cycle list )
            ( cycles2 : cycle list ) : cycle =
    match cycles1, cycles2 with
    | [], [] -> c_euler
    | [], _ -> iterations c_euler cycles2 []
    | c :: cycles1, _ ->
        let s = recherche_sommet_commun c c_euler in
        if s = ( -1  ) then iterations c_euler cycles1 ( c :: cycles2 )
        else iterations ( fusion c_euler c s ) cycles1 cycles2 in
    match cycles with
    | [] -> []
    | c :: cycles -> iterations c cycles [] in

fusion_liste !cycles
  end
;;

let cycle_hamiltonien ( g : graphe ) : cycle =
  let rec parcours_exhaustif
      ( s : sommet )
      ( visite : bool array )
      ( nb_visite : int )
      ( cycle : cycle ) : cycle =
    if nb_visite = ( Array.length g )
    then ( if List.mem 0 g.(s) then ( 0 :: cycle ) else [] )
    else begin
      let rec parcours_voisins = function
        | [] -> []
        | v :: voisins ->
          if visite.(v) then parcours_voisins voisins
          else begin
            let visite_v = Array.copy visite in
            visite_v.(v) <- true ;
            let c =
              parcours_exhaustif v visite_v ( nb_visite + 1 ) ( v :: cycle ) in
            if c = [] then parcours_voisins voisins else c
          end in
      parcours_voisins g.(s)
    end in
  let visite = Array.make ( Array.length g ) false in
  visite.(0) <- true;
  parcours_exhaustif 0 visite 1 [0]
;;
